function [ H, sigma ] = estimEntropy( S, sigma, restart )
% function [ H, sigma ] = estimEntropy( S, sigma )
%    Estimate entropy of a continuous time series given a finite sample,
%    using Parzen method, according to [Hlavackova-Schindler_PhysicsReports_2007].
%    If sigma is not supplied, it is estimated using EM algorithm.
	
	if nargin < 3
		restart = false;
	end

	maxPts = 100; % 100
	d = size(S, 1);
	len = size(S,2);
	len = min(len, max(2*d, maxPts)); % limit, BUT at least 2*D
	S = S(:, end-len+1:end);
	
	minPts = size(S,2)+1;
	[S, prevIntDiff] = interpolatePoints(S, minPts); % if there's too few points, add some interpolated
	prevIntDiff = false; % we interpolate always anyway... so the number of int. points is 1 and doesn't change
	
	if nargin < 2 || isempty(sigma) || prevIntDiff
		[sigma, converged] = sigmaEM(S, [], 1e-4, 100);
	elseif restart
		[sigma, converged] = sigmaEM(S, sigma, 1e-4, 50);
	else
		converged = true;
	end
	
	H = 0;
	for i = 1:size(S, 2)
		y = S(:,i); % the vector where we measure the probability
		T = [S(:,1:i-1) S(:, i+1:end)]; % the other vectors for KDE (leave-one-out)
		logp = log2(KDEprob(y, T, sigma));
		if ~isinf(logp)
			H = H + logp;
		end
	end
	
	H = - H / size(S, 2); % we don't add log|T| here, because the KDEprob function returns the pdf already normalised
	
	if ~converged
% 		fprintf('|Not converged!|');
		sigma = [];
	end
	
end

