'''
Script to train a gaussian process for regressing points on an objects surface.

command line args:
-i input filepath
-b bias level
-hyper train hyperparameters (0 is just recondition, 1 is full retrain, 2 is retrain only noise level)
-d num_dimensions
-n num_pts
-u flag to input point uncertainty
-v verbose
-f filename to save out GP to

return codes:
0    success
1    invalid csv file dimensionality
2    File IO failure

Created on Jun 17, 2013

@author: Simon
'''

import argparse
from numpy import *
import GPy
import cPickle
import sys

USE_BIAS=True
USE_WHITE=False
USE_EXP=False
USE_RBF=True
USE_LINEAR=False

verbose_flag=False

def create_GP(vals,dims,uncert_flag):
    kernel=None
    kernelSet = False
    if USE_EXP:
        if kernelSet:
            kernel=kernel+GPy.kern.exponential(dims)
        else:
            kernel=GPy.kern.exponential(dims)
        kernelSet = True
    if USE_RBF:
        if kernelSet:
            kernel=kernel+GPy.kern.rbf(dims, ARD = True)
        else:
            kernel=GPy.kern.rbf(dims, ARD = True)
        kernelSet = True
    if USE_LINEAR:
        if kernelSet:
            kernel=kernel+GPy.kern.linear(dims)
        else:
            kernel=GPy.kern.linear(dims)
        kernelSet = True
    if USE_BIAS:
        if kernelSet:
            kernel = kernel+GPy.kern.bias(dims)
        else:
            kernel = GPy.kern.bias(dims)
        kernelSet = True
    if USE_WHITE:
        if kernelSet:
            kernel=kernel+GPy.kern.white(dims)
        else:
            kernel=GPy.kern.white(dims)
        kernelSet = True
    if uncert_flag==1:
        if USE_EXP:
            raise Exception("Can't use uncertainties with Exponential kernel, currently unimplemented in GPy")
        Y= vals[:,-1]
        Y=Y[:,None]
        X=vals[:,0:dims]
        U=vals[:,dims:-1]
    else:
        Y= vals[:,-1]
        Y=Y[:,None]
        X=vals[:,0:-1]
        U=None
    num_pts=size(X,0)
    if num_pts<200 or True:
        model=GPy.models.GPRegression(X,Y, kernel)
    else:
        model=GPy.models.SparseGPRegression(X,Y, kernel,num_inducing=min(50,num_pts/10),X_variance=U)
    return model

def copy_gp_hyperparam(new_model,old_model,param_name):
    """Copies the value of a hyperparameter from one model to another"""
    try:
        new_model[param_name]=old_model[param_name]
    except:
        vprint(["Parameter ",param_name," Was not in old model, leaving as default value."])

def copy_all_gp_hyperparams(new_model,old_model):
    """Copies the value of all hyperparameters from one model to another"""
    if USE_EXP:
        copy_gp_hyperparam(new_model,old_model,"exp_variance")
        copy_gp_hyperparam(new_model,old_model,"exp_lengthscale")
    if USE_RBF:
        copy_gp_hyperparam(new_model,old_model,"rbf_variance")
        copy_gp_hyperparam(new_model,old_model,"rbf_lengthscale")
    if USE_BIAS:
        copy_gp_hyperparam(new_model,old_model,"bias")
    if USE_WHITE:
        copy_gp_hyperparam(new_model,old_model,"white_variance")
    if USE_LINEAR:
        copy_gp_hyperparam(new_model,old_model,"linear_variance")

def constrain_all_kernel_hyperparams(new_model):
    """applies constrain_fixed to all hyperparameters from ther kernels we are using (but not to inducing points and noise level"""
    if USE_EXP:
        new_model.constrain_fixed("exp_variance")
        new_model.constrain_fixed("exp_lengthscale")
    if USE_RBF:
        new_model.constrain_fixed("rbf_variance")
        new_model.constrain_fixed("rbf_lengthscale")
    if USE_BIAS:
        new_model.constrain_fixed("bias")
    if USE_LINEAR:
        new_model.constrain_fixed("linear_variance")

def train_model(vals,bias,dims,uncert_flag,old_model=None):
    """Take a numpy array where final column is radius and rest of matrix is normalised 3d points for training.
    If an old model is passed, it's hyperparameters provide initialisation for the optimisation"""
    model=create_GP(vals,dims,uncert_flag)
    if not old_model==None:
        copy_all_gp_hyperparams(model,old_model)
    else:
        try:
            model['.*length']=1
        except:
            vprint(verbose_flag, [".*length not found, probably using kernels without such thing..."])
    model.ensure_default_constraints()
    if USE_BIAS:
        if not bias==0:
            model['.*bias']=bias
            model.constrain_fixed("bias")
    vprint(['beginning optimization with initial gp',model])
    model.optimize_restarts(max_iters=100,ftol=1e-2,messages=verbose_flag, num_restarts = 5,parallel=True, robust = True) #//SPEED: max_iters might probably be lowered
    return model

def train_noise_level(vals,old_model,dims,uncert_flag):
    """Retrain only the noise_variance hyperparameter of an existing model, for new data points"""
    model=create_GP(vals,dims,uncert_flag)

    #copy hyperparameters from old_model
    copy_all_gp_hyperparams(model,old_model)

    #force all but noise and iips to be fixed
    constrain_all_kernel_hyperparams(model)
    noise=old_model["noise_variance"]
    noise_range=0.3
    model.constrain_bounded("noise_variance",(1-noise_range)*noise,(1+noise_range)*noise)
    if USE_WHITE:
        noisew=old_model["white_variance"]
        model.constrain_bounded("white_variance",(1-noise_range)*noisew,(1+noise_range)*noisew)
    vprint(['beginning optimization with initial gp',model])
    model.optimize(max_iters=100,ftol=1e-2,messages=verbose_flag)

    #maybe only accept if log likelihood goes up?
    return model

def recondition(vals,old_model,dims,uncert_flag):
    """Retrain only the noise_variance hyperparameter of an existing model, for new data points"""
    model=create_GP(vals,dims,uncert_flag)

    #copy hyperparameters from old_model
    copy_all_gp_hyperparams(model,old_model)

    #constrain everything
    constrain_all_kernel_hyperparams(model)
    model.constrain_fixed("noise_variance")
    if USE_WHITE:
        model.constrain_fixed("white_variance")
    vprint(['beginning optimization with initial gp',model])
    model.optimize(max_iters=100,ftol=1e-2,messages=verbose_flag)

    #maybe only accept if log likelihood goes up?
    return model

def vprint( text):
    if verbose_flag==1:
        for item in text:
            print "verbose: ",item
        print "\n"

if __name__ == '__main__':

    random.seed(100) #if numpy isn't seeded at all, the output is always the same anyway, but lets explicitly seed it with a constant

    #read command line args
    parser = argparse.ArgumentParser(description='Script to train a gaussian process for regressing points on an objects surface.')
    parser.add_argument('-f',nargs=1,type=str,required=True,help="path to the storage file")
    parser.add_argument('-i',nargs=1,type=str,required=True,help="path to the input CSV file")
    parser.add_argument('-b',nargs=1,type=float,default=[0],help="the level of bias, i.e. general radius in 3d")
    parser.add_argument('-hyper',nargs=1,type=int,choices=[0,1,2],default=[0],help="flag for retraining of hyperparameters (0 is just recondition, 1 is full retrain, 2 is retrain only noise level)")
    parser.add_argument('-d',nargs=1,type=int,default=[0],help="number of dimensions for input points")
    parser.add_argument('-n',nargs=1,type=int,default=[0],help="number of input points")
    parser.add_argument('-v',nargs=1,type=int,choices=[0,1],default=[0],help="print out verbose information about training")
    parser.add_argument('-u',nargs=1,type=int,choices=[0,1],default=[0],help="flag to enable input of point uncertainties")
    args=parser.parse_args()

    verbose_flag=args.v[0]
    if verbose_flag:
        import time
        start=time.time()
    vprint(["Command line arguments were",args])

    #read csv file, should be n by d+1
    vprint(["Loading csv file from ",args.i[0]])
    try:
        csvfile=open(args.i[0])
    except:
        print "Failed to open CSV file at ",args.i[0]
        sys.exit(2)

    vals=loadtxt(csvfile, float, delimiter=',', ndmin=2)
    vprint(["Loaded matrix had shape",shape(vals)])

    if not size(vals,0) == args.n[0]:
        print "ERROR: Command line arguments specified ",args.n[0],"points but read",size(vals,0),"from file at",args.i
        sys.exit(1)
    input_dims=args.d[0]+1
    if args.u[0]==1:
        input_dims=2*args.d[0]+1
    if not size(vals,1) == input_dims:
        print "ERROR: Command line arguments specified ",input_dims,"dimensions but read",size(vals,1),"from file at",args.i,"(should be 1 more than cmd line)"
        sys.exit(1)

    #either load model or train model
    if args.hyper[0]==1:
        vprint(["Completely retraining hyperparameters "])
        #try to load anyway, as an initialisation
        try:
            modelfile=open(args.f[0])
            old_m=cPickle.load(modelfile)
            vprint(["loaded previous file as initial hyperparameters "])
        except:
            old_m=None
            vprint(["Failed to load previous file, initialising to 1s "])
        m=train_model(vals,args.b[0],args.d[0],args.u[0], old_m)
        vprint(["Newly trained model had values ",m])
    else:
        vprint(["Loading existing model file from "+args.f[0]])
        try:
            modelfile=open(args.f[0])
        except:
            print "Command line arguments specified no hyperparameter training, but failed to open model file at /"+args.f[0]
            sys.exit(2)
        #load previous model + update bias
        old_m=cPickle.load(modelfile)
        vprint(["Loaded model had values",old_m])
#        vprint(["Setting bias to ",args.b])
#        m["bias"]=args.b

        if args.hyper[0]==2:
            vprint(["Retraining only noise level"])
            m=train_noise_level(vals,old_m,args.d[0],args.u[0])
        else:
            vprint(["Reconditioning without updating hyperparameters"])
            m=recondition(vals,old_m,args.d[0],args.u[0])
        vprint(["New model",m])
    
    #write out model file
    s=cPickle.dumps(m)
    f=open(args.f[0],"w")
    f.write(s)

    if verbose_flag:
        print "Program time=",time.time()-start,"seconds"
    sys.exit(0)

